;    Random number generator from hardware noise source
;    Copyright © 2009 Jeff Epler <jepler@unpythonic.net>
;
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#define CRC_L r18
#define CRC_H r19
#define TMP   r20
#define CNT   r21
#define OTHER r22

#include <avr/io.h>

; Perform one step of the random number generator: get one bit from the
; noise source and use it as the next bit of a CRC16 calculation.  Note
; that STEP takes a constant amount of time to execute regardless of the
; noise bit read and the old contents of CRC_L.  After a large number of
; STEPs, each bit of noise has a chance to affect each bit of CRC_H:CRC_L.
.macro STEP8
eor TMP, TMP
sbic _SFR_IO_ADDR(PINC), 0
ori TMP, 1
sbic _SFR_IO_ADDR(PINC), 0
ori TMP, 2
sbic _SFR_IO_ADDR(PINC), 0
ori TMP, 4
sbic _SFR_IO_ADDR(PINC), 0
ori TMP, 8
sbic _SFR_IO_ADDR(PINC), 0
ori TMP, 16
sbic _SFR_IO_ADDR(PINC), 0
ori TMP, 32
sbic _SFR_IO_ADDR(PINC), 0
ori TMP, 64
sbic _SFR_IO_ADDR(PINC), 0
ori TMP, 128

eor CRC_L, TMP
mov TMP, CRC_L
swap CRC_L
andi CRC_L, 0xf0
eor CRC_L, TMP
mov TMP, CRC_H
mov CRC_H, CRC_L
swap CRC_L
andi CRC_L, 0xf
eor TMP, CRC_L
lsr CRC_L
eor CRC_H, CRC_L
eor CRC_L, CRC_H
lsl CRC_L
lsl CRC_L
lsl CRC_L
eor CRC_L, TMP
.endm

.globl main
main:
; 2 megabaud, transmit only
	ldi TMP, 0
	sts _SFR_MEM_ADDR(UBRR0L), TMP
	sts _SFR_MEM_ADDR(UBRR0H), TMP
	ldi TMP, (1<<U2X0)
	sts _SFR_MEM_ADDR(UCSR0A), TMP
	ldi TMP, (1<<TXEN0)
	sts _SFR_MEM_ADDR(UCSR0B), TMP

; prime the CRC with 8*256 bits of noise
	ldi CNT, 0
initloop:
	STEP8
	dec CNT
	brne initloop
	mov OTHER, CRC_H

; repeatedly collect 64 bits of noise and output 16 bits of randomness.
; UDR0 is guaranteed empty each time it's written, because of the number
; of cycles it takes to STEP 32 times.  The first time through OTHER is
; loaded with the high bits of the CRC after 256 STEPs, so even the first
; bits output after reset should be suitably random.
loop:
.rept 4
	STEP8
.endr
	sts _SFR_MEM_ADDR(UDR0), OTHER
.rept 4
	STEP8
.endr
	sts _SFR_MEM_ADDR(UDR0), CRC_L
	mov OTHER, CRC_H
	rjmp loop
